from datahugger.services import ArXivDataset
from datahugger.services import DataDryadDataset
from datahugger.services import DataOneDataset
from datahugger.services import DataverseDataset
from datahugger.services import DjehutyDataset
from datahugger.services import DSpaceDataset
from datahugger.services import FigShareDataset
from datahugger.services import GitHubDataset
from datahugger.services import HuggingFaceDataset
from datahugger.services import MendeleyDataset
from datahugger.services import OSFDataset
from datahugger.services import PangaeaDataset
from datahugger.services import ZenodoDataset

# fast lookup
SERVICES_NETLOC = {
    "arxiv.org": ArXivDataset,
    "zenodo.org": ZenodoDataset,
    "github.com": GitHubDataset,
    "datadryad.org": DataDryadDataset,
    "huggingface.co": HuggingFaceDataset,
    "osf.io": OSFDataset,
    "data.mendeley.com": MendeleyDataset,
    # Figshare download
    "figshare.com": FigShareDataset,
    # Djehuty
    "data.4tu.nl": DjehutyDataset,
    # DataOne repositories
    "arcticdata.io": DataOneDataset,
    "knb.ecoinformatics.org": DataOneDataset,
    "data.pndb.fr": DataOneDataset,
    "opc.dataone.org": DataOneDataset,
    "portal.edirepository.org": DataOneDataset,
    "goa.nceas.ucsb.edu": DataOneDataset,
    "data.piscoweb.org": DataOneDataset,
    "adc.arm.gov": DataOneDataset,
    "scidb.cn": DataOneDataset,
    "data.ess-dive.lbl.gov": DataOneDataset,
    "hydroshare.org": DataOneDataset,
    "ecl.earthchem.org": DataOneDataset,
    "get.iedadata.org": DataOneDataset,
    "usap-dc.org": DataOneDataset,
    "iys.hakai.org": DataOneDataset,
    "doi.pangaea.de": PangaeaDataset,
    "rvdata.us": DataOneDataset,
    "sead-published.ncsa.illinois.edu": DataOneDataset,
    # DataVerse repositories (extracted from re3data)
    "dataverse.acg.maine.edu": DataverseDataset,
    "dataverse.icrisat.org": DataverseDataset,
    "datos.pucp.edu.pe": DataverseDataset,
    "datos.uchile.cl": DataverseDataset,
    "opendata.pku.edu.cn": DataverseDataset,
    "www.march.es": DataverseDataset,
    "www.murray.harvard.edu": DataverseDataset,
    "abacus.library.ubc.ca": DataverseDataset,
    "ada.edu.au": DataverseDataset,
    "adattar.unideb.hu": DataverseDataset,
    "archive.data.jhu.edu": DataverseDataset,
    "borealisdata.ca": DataverseDataset,
    "dados.ipb.pt": DataverseDataset,
    "dadosdepesquisa.fiocruz.br": DataverseDataset,
    "darus.uni-stuttgart.de": DataverseDataset,
    "data.aussda.at": DataverseDataset,
    "data.cimmyt.org": DataverseDataset,
    "data.fz-juelich.de": DataverseDataset,
    "data.goettingen-research-online.de": DataverseDataset,
    "data.inrae.fr": DataverseDataset,
    "data.scielo.org": DataverseDataset,
    "data.sciencespo.fr": DataverseDataset,
    "data.tdl.org": DataverseDataset,
    "data.univ-gustave-eiffel.fr": DataverseDataset,
    "datarepositorium.uminho.pt": DataverseDataset,
    "datasets.iisg.amsterdam": DataverseDataset,
    "dataspace.ust.hk": DataverseDataset,
    "dataverse.asu.edu": DataverseDataset,
    "dataverse.cirad.fr": DataverseDataset,
    "dataverse.csuc.cat": DataverseDataset,
    "dataverse.harvard.edu": DataverseDataset,
    "dataverse.iit.it": DataverseDataset,
    "dataverse.ird.fr": DataverseDataset,
    "dataverse.lib.umanitoba.ca": DataverseDataset,
    "dataverse.lib.unb.ca": DataverseDataset,
    "dataverse.lib.virginia.edu": DataverseDataset,
    "dataverse.nl": DataverseDataset,
    "dataverse.no": DataverseDataset,
    "dataverse.openforestdata.pl": DataverseDataset,
    "dataverse.scholarsportal.info": DataverseDataset,
    "dataverse.theacss.org": DataverseDataset,
    "dataverse.ucla.edu": DataverseDataset,
    "dataverse.unc.edu": DataverseDataset,
    "dataverse.unimi.it": DataverseDataset,
    "dataverse.yale-nus.edu.sg": DataverseDataset,
    "dorel.univ-lorraine.fr": DataverseDataset,
    "dvn.fudan.edu.cn": DataverseDataset,
    "edatos.consorciomadrono.es": DataverseDataset,
    "edmond.mpdl.mpg.de": DataverseDataset,
    "heidata.uni-heidelberg.de": DataverseDataset,
    "lida.dataverse.lt": DataverseDataset,
    "mxrdr.icm.edu.pl": DataverseDataset,
    "osnadata.ub.uni-osnabrueck.de": DataverseDataset,
    "planetary-data-portal.org": DataverseDataset,
    "qdr.syr.edu": DataverseDataset,
    "rdm.aau.edu.et": DataverseDataset,
    "rdr.kuleuven.be": DataverseDataset,
    "rds.icm.edu.pl": DataverseDataset,
    "recherche.data.gouv.fr": DataverseDataset,
    "redu.unicamp.br": DataverseDataset,
    "repod.icm.edu.pl": DataverseDataset,
    "repositoriopesquisas.ibict.br": DataverseDataset,
    "research-data.urosario.edu.co": DataverseDataset,
    "researchdata.cuhk.edu.hk": DataverseDataset,
    "researchdata.ntu.edu.sg": DataverseDataset,
    "rin.lipi.go.id": DataverseDataset,
    "ssri.is": DataverseDataset,
    "trolling.uit.no": DataverseDataset,
    "www.sodha.be": DataverseDataset,
    "www.uni-hildesheim.de": DataverseDataset,
}

# regexp lookup
SERVICES_NETLOC_REGEXP = {
    r".*\/articles\/.*\/.*\/\d+": FigShareDataset,
    r".*\/handle\/\d+\/\d+": DSpaceDataset,
    r".*\/dataset\.xhtml\?persistentId\=.*": DataverseDataset,
}

# add keys in lower-case for fast case-insensitive lookups
RE3DATA_SOFTWARE = {
    "dataverse": DataverseDataset,
    "dspace": DSpaceDataset,
    # "CKAN": CKANDataset,  # Hits on re3data 2022-09-02: (89)
    # "MySQL": MySQLDataset,  # Hits on re3data 2022-09-02: (86)
    # "Fedora": FedoraDataset,  # Hits on re3data 2022-09-02: (43)
    # "EPrints": EPrintsDataset,  # Hits on re3data 2022-09-02: (34)
    # "Nesstar": NesstarDataset,  # Hits on re3data 2022-09-02: (19)
    # "DigitalCommons": DigitalCommonsDataset,  # Hits on re3data 2022-09-02: (4)
    # "eSciDoc": eSciDocDataset,  # Hits on re3data 2022-09-02: (3)
    # "Opus": OpusDataset,  # Hits on re3data 2022-09-02: (2)
    # "dLibra": dLibraDataset,  # Hits on re3data 2022-09-02: (2)
}



import re
import xml.etree.ElementTree as ET
from pathlib import Path
from urllib.parse import urlparse

import requests
import requests_cache


def _is_url(s: str) -> bool:
    """Check if the string is a URL.

    Arguments
    ---------
    s: str
        The url to check.

    Returns
    -------
    bool:
        Boolean that indicates whether the string is url."""

    return urlparse(s).netloc != ""


def _get_url(s: str) -> str:
    return s if isinstance(s, str) else s.url


def _format_filename(s, len_s=35) -> str:
    # README_Pfaller_Robinson_2022_Global_Sea_Turtle_Epibiont_Database.txt

    if len_s is None:
        return s

    s = str(s)

    if len(s) <= len_s:
        return s.ljust(len_s, " ")

    len_suffixes = len("".join(Path(s).suffixes))

    return s[0 : (len_s - (len_suffixes + 5))] + "[...]" + "".join(Path(s).suffixes)


def get_id_from_url(regexp, url):
    match = re.search(regexp, url)

    if match.group(1):
        return match.group(1)


def get_datapublisher_from_doi(doi):
    """Get the publisher from the DOI.

    Arguments
    ---------
    doi: str
        The DOI to find the publisher for.

    Returns
    -------
    str:
        The publisher.

    """

    r = requests.get(f"https://api.datacite.org/dois/{doi}")
    r.raise_for_status()

    record = r.json()

    return record["data"]["attributes"]["publisher"]


def get_re3data_repositories(
    url="https://www.re3data.org/api/v1/repositories", expire_after=3600
):
    # use cached version here
    session = requests_cache.CachedSession(
        "datahugger_cache",
        expire_after=expire_after,
        backend="filesystem",
        use_cache_dir=True,
    )
    r = session.get(url)
    r.raise_for_status()

    tree = ET.fromstring(r.content)

    for node in tree:
        yield {elem.tag: elem.text for elem in node if not elem.tag == "link"}


def get_re3data_repository(re3data_id):
    namespaces = {"r3d": "http://www.re3data.org/schema/2-2"}
    r = requests.get(f"https://www.re3data.org/api/v1/repository/{re3data_id}")
    r.raise_for_status()

    tree = ET.fromstring(r.content)

    return (
        tree[0]
        .find("r3d:software", namespaces)
        .find("r3d:softwareName", namespaces)
        .text
    )



from pathlib import Path

import pytest

import datahugger
from datahugger.utils import _is_url

TESTS_URLS = [
    # Zenodo
    ("10.5281/zenodo.6625880", "README.md"),
    # Dataverse
    (
        "https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/DVN/KBHLOD",
        "tutorial1.py",
    ),
    ("https://doi.org/10.7910/DVN/KBHLOD", "tutorial1.py"),
    ("https://hdl.handle.net/10622/NHJZUD", "ERRHS_7_01_data_1795.tab"),
    # Dataverse single file
    ("10.7910/DVN/HZBYG7/RQ26H2", "Table 2.do"),
    # Figshare
    ("https://doi.org/10.6084/m9.figshare.8851784.v1", "cross_year_data2.csv"),
    (
        "https://figshare.com/articles/dataset/Long-term_behavioral_repeatability_in_wild_adult_and_captive_juvenile_turtles_implications_for_personality_development/8851784",
        "cross_year_data2.csv",
    ),
    (
        "https://doi.org/10.15131/shef.data.22010159.v2",
        "ScHARR QUIT evaluation statistical and health economic analysis plan.pdf",
    ),
    # Djehuty
    ("https://doi.org/10.4121/21989216.v1", "README.txt"),
    # Dryad
    (
        "https://datadryad.org/stash/dataset/doi:10.5061/dryad.31zcrjdm5",
        "ReadmeFile.txt",
    ),
    ("https://doi.org/10.5061/dryad.31zcrjdm5", "ReadmeFile.txt"),
    # OSF
    ("https://osf.io/kq573/", "nest_area_data.xlsx"),
    ("https://doi.org/10.17605/OSF.IO/KQ573", "nest_area_data.xlsx"),
    ("https://doi.org/10.17605/OSF.IO/9X6CA", "jobs.sh"),
    # Mendeley
    (
        "https://doi.org/10.17632/p6wmtv6t5g.2",
        "READMI Stranding Sea Turtle records.pdf",
    ),
    ("https://doi.org/10.17632/p6wmtv6t5g", "READMI Stranding Sea Turtle records.pdf"),
    # Dataone
    ("https://doi.org/10.18739/A2KH0DZ42", "2012F_Temperature_Data.csv"),
    # DSpace
    ("https://uhra.herts.ac.uk/handle/2299/26087", "pdf.pdf"),
    (
        "https://repositorioinstitucional.ceu.es/handle/10637/2741",
        "Aquaporin_1_JAMartin_et_al_MedSport_2009.pdf",
    ),
    # huggingface
    # ("10.57967/hf/0034", "test.csv"),
    # Pangaea
    ("https://doi.org/10.1594/PANGAEA.954547", "Gubbio_age.tab"),
    ("https://doi.pangaea.de/10.1594/PANGAEA.954543", "AA_age.tab"),
]


def test_url_checker():
    assert not _is_url("82949")
    assert _is_url("https://doi.org/10.5281/zenodo.6614829")


@pytest.mark.parametrize("url_or_id,output_file", TESTS_URLS)
def test_load_dyno(url_or_id, output_file, tmpdir):
    """Load repository with the generic loader."""
    datahugger.get(url_or_id, tmpdir)

    assert Path(tmpdir, output_file).exists()


@pytest.mark.parametrize(
    "url_or_id",
    [
        ("10.5281/zenodo.6614829"),
        ("https://zenodo.org/record/6614829"),
        ("https://doi.org/10.5281/zenodo.6614829"),
    ],
)
def test_load_zenodo_6614829(url_or_id, tmpdir):
    datahugger.get(url_or_id, tmpdir, max_file_size=1e6)

    assert Path(tmpdir, "quasiperiod.m").exists()
    assert not Path(tmpdir, "quasiperiod.txt.gz").exists()


def test_load_github_cbsodata(tmpdir):
    datahugger.get("https://github.com/j535d165/cbsodata", tmpdir)


def test_info_without_loading(tmpdir):
    dh_get = datahugger.get("https://osf.io/wdzh5/", output_folder=".", print_only=True)

    dh_info = datahugger.info("https://osf.io/wdzh5/")

    assert dh_get.dataset.files == dh_info.files


def test_huggingface(tmpdir):
    datahugger.get(
        "https://huggingface.co/datasets/wikitext",
        tmpdir,
        params={"name": "wikitext-2-v1"},
    )


def test_huggingface_without_params(tmpdir):
    with pytest.raises(ValueError):
        datahugger.get("https://huggingface.co/datasets/wikitext", tmpdir)



"""Powerful datahugger command line interface

Example:

    python -m datahugger https://doi.org/10.5061/dryad.31zcrjdm5 $(mktemp -d)

"""

import argparse
import logging

from datahugger import __version__
from datahugger import get
from datahugger.exceptions import DOIError


def print_red(s):
    print(f"\u001b[31m{s}\u001b[0m")


def print_green(s):
    print(f"\u001b[32m{s}\u001b[0m")


class KVAppendAction(argparse.Action):
    def __call__(self, parser, args, values, option_string=None):
        try:
            (k, v) = values[0].split("=", 2)
        except ValueError as err:
            raise argparse.ArgumentError(
                self, f"'{values[0]}' is not a valid key-value pair (key=value)"
            ) from err

        d = getattr(args, self.dest) or {}
        d[k] = v
        setattr(args, self.dest, d)


def main():
    parser = argparse.ArgumentParser(
        prog="datahugger",
        description="One downloader for all scientific data.",
    )
    parser.add_argument(
        "url_or_doi",
        help="An URL or DOI to scientific data repository.",
    )

    parser.add_argument(
        "output_dir",
        help="The dir to store the output to.",
    )

    parser.add_argument(
        "--max-file-size",
        default=None,
        type=int,
        help="Skip files larger than this size. Might not work for all services.",
    )

    parser.add_argument(
        "-f", "--force-download", dest="force_download", action="store_true"
    )
    parser.set_defaults(force_download=False)

    parser.add_argument("--no-unzip", dest="unzip", action="store_false")
    parser.set_defaults(unzip=True)

    parser.add_argument("--no-progress", dest="progress", action="store_false")
    parser.set_defaults(progress=True)

    parser.add_argument(
        "-d",
        "--dry-run",
        dest="print_only",
        action="store_true",
        help="Only print the files to download without downloading them.",
    )

    parser.add_argument(
        "--log-level",
        default="WARNING",
        help="Python based log levels. Default: WARNING.",
    )

    parser.add_argument(
        "-p",
        "--param",
        nargs=1,
        action=KVAppendAction,
        dest="params",
        help="Add key=value params to pass to the downloader. "
        "May appear multiple times.",
    )

    # version
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )

    args, _ = parser.parse_known_args()

    logging.basicConfig(level=args.log_level)

    try:
        # Start downloading
        get(
            args.url_or_doi,
            args.output_dir,
            max_file_size=args.max_file_size,
            force_download=args.force_download,
            unzip=args.unzip,
            progress=args.progress,
            print_only=args.print_only,
            params=args.params,
        )

    except DOIError as doi_err:
        # raise error when log level is DEBUG
        if logging.DEBUG == logging.root.level:
            raise doi_err
        else:
            print_red(f"Error: {doi_err}")
            print("")
            print("Check if your DOI is correct at doi.org. Is the DOI valid?")
            print("Please request support in the issue tracker:")
            print("")
            print("\thttps://github.com/J535D165/datahugger/issues/new/choose")
            print()
            exit(1)

    except Exception as err:
        # raise error when log level is DEBUG
        if logging.DEBUG == logging.root.level:
            raise err
        else:
            print_red(f"Error: {err}")
            print("")
            print("Do you think this is a data repository that needs to be supported?")
            print("Please request support in the issue tracker:")
            print("")
            print("\thttps://github.com/J535D165/datahugger/issues/new/choose")
            print()
            exit(1)

    if args.progress:
        print_green("Dataset successfully downloaded.")


if __name__ == "__main__":
    main()



import io
import re
import xml.etree.ElementTree as ET
import zipfile
from pathlib import Path
from typing import Union
from urllib.parse import quote
from urllib.parse import urlparse

import requests
from jsonpath_ng.jsonpath import Fields
from jsonpath_ng.jsonpath import Slice

from datahugger.base import DatasetDownloader
from datahugger.utils import _get_url


class ArXivDataset(DatasetDownloader):
    """Downloader for ArXiv publication."""

    REGEXP_ID = r"https://arxiv\.org/abs/(?P<record_id>.*)"

    @property
    def files(self):
        return [
            {
                "link": f"https://arxiv.org/pdf/{self._params['record_id']}.pdf",
                "name": self._params["record_id"].split("/")[-1] + ".pdf",
                "size": None,
                "hash": None,
                "hash_type": None,
            }
        ]


class DataverseDataset(DatasetDownloader):
    """Downloader for Dataverse repository."""

    REGEXP_ID = r"(?P<type>dataset|file)\.xhtml\?persistentId=(?P<record_id>.*)"

    # paths to file attributes
    ATTR_NAME_JSONPATH = "filename"
    ATTR_SIZE_JSONPATH = "filesize"
    ATTR_HASH_JSONPATH = "md5"
    ATTR_HASH_TYPE_VALUE = "md5"

    @property
    def API_URL_META(self):
        if self._params.get("version", None):
            v = self._params["version"]
        else:
            v = ":latest-published"

        if self._params.get("type", None) == "file":
            return "{base_url}/api/files/:persistentId/?persistentId={record_id}"
        else:
            return (
                "{base_url}/api/datasets/:persistentId/versions/"
                f"{v}/?persistentId={{record_id}}"
            )

    @property
    def META_FILES_JSONPATH(self):
        if self._params.get("type", None) == "file":
            return "data.dataFile"
        else:
            return "data.files[*].dataFile"

    def _get_attr_link(self, record, base_url=None):
        return f"{base_url}/api/access/datafile/{record['id']}"


class DataDryadDataset(DatasetDownloader):
    """Downloader for DataDryad repository."""

    REGEXP_ID = r"datadryad\.org[\:]*[43]{0,3}\/stash\/dataset\/doi:(?P<record_id>.*)"

    # the base entry point of the REST API
    API_URL = "https://datadryad.org/api/v2"

    # the files and metadata about the dataset
    API_URL_META = "{api_url}{record_id}/files/osfstorage/?format=jsonapi"
    META_FILES_JSONPATH = (
        Fields("_embedded").child(Fields("stash:files")).child(Slice())
    )

    # paths to file attributes
    ATTR_NAME_JSONPATH = "path"
    ATTR_SIZE_JSONPATH = "size"

    @property
    def API_URL_META(self):
        doi_safe = quote(f"doi:{self._params['record_id']}", safe="")
        dataset_metadata_url = self.API_URL + "/datasets/" + doi_safe

        res = requests.get(dataset_metadata_url)
        res.raise_for_status()
        dataset_metadata = res.json()

        # get the latest version of the dataset
        latest_version = dataset_metadata["_links"]["stash:version"]["href"]
        return f"https://datadryad.org{latest_version}/files"

    def _get_attr_link(self, record, base_url):
        return base_url + record["_links"]["stash:file-download"]["href"]


class DataOneDataset(DatasetDownloader):
    """Downloader for DataOne repositories."""

    REGEXP_ID = r"view/doi:(?P<record_id>.*)"

    # the base entry point of the REST API
    API_URL = "https://cn.dataone.org/cn/v2/object/"

    @property
    def files(self):
        if hasattr(self, "_files"):
            return self._files

        doi_safe = quote(f"doi:{self._params['record_id']}", safe="")

        res = requests.get(self.API_URL + doi_safe)
        res.raise_for_status()
        meta_tree = ET.fromstring(res.content)

        x = []
        for data_elem in meta_tree.find("dataset"):
            if data_elem.tag in ["otherEntity", "dataTable"]:
                x.append(
                    {
                        "link": data_elem.find(
                            "./physical/distribution/online/url[@function='download']"
                        ).text,
                        "name": data_elem.find("entityName").text,
                        "size": data_elem.find("./physical/size").text,
                        "hash": None,
                        "hash_type": None,
                    }
                )

        self._files = x
        return self._files


class PangaeaDataset(DatasetDownloader):
    """Downloader for PangaeaDataset repository."""

    REGEXP_ID = r"doi\.pangaea\.de/(?P<record_id>.*)"

    # the base entry point of the REST API
    API_URL = "https://doi.pangaea.de/"

    @property
    def files(self):
        # get the difference between collection and file
        r = requests.get(
            f"{self.API_URL}{self._params['record_id']}?format=metadata_jsonld"
        )
        r.raise_for_status()
        dists = r.json()["distribution"]

        if isinstance(dists, dict):
            dists = [dists]

        files = []
        for d in dists:
            if d["encodingFormat"] in ["text/tab-separated-values", "application/zip"]:
                r_filename = requests.head(d["contentUrl"])
                content_d = r_filename.headers["content-disposition"]

                files.append(
                    {
                        "link": d["contentUrl"],
                        "name": re.findall("filename=(.+)", content_d)[0],
                        "size": None,
                        "hash": None,
                        "hash_type": None,
                    }
                )

        return files


class DSpaceDataset(DatasetDownloader):
    """Downloader for DSpaceDataset repositories."""

    REGEXP_ID = r"handle/(?P<record_id>\d+\/\d+)"

    # paths to file attributes
    ATTR_KIND_JSONPATH = "attributes.kind"

    ATTR_FILE_LINK_JSONPATH = "link"

    ATTR_NAME_JSONPATH = "name"
    ATTR_SIZE_JSONPATH = "sizeBytes"
    ATTR_HASH_JSONPATH = "checkSum.checkSumAlgorithm"
    ATTR_HASH_TYPE_VALUE = "checkSum.value"

    def _get_attr_link(self, record, base_url):
        return base_url + record["retrieveLink"]

    @property
    def API_URL_META(self):
        uri = urlparse(_get_url(self.resource))
        base_url = uri.scheme + "://" + uri.netloc

        handle_id_url = f"{base_url}/rest/handle/{self._params['record_id']}"
        res = requests.get(handle_id_url)
        res.raise_for_status()

        return base_url + res.json()["link"] + "/bitstreams"


class FigShareDataset(DatasetDownloader):
    """Downloader for FigShare repository."""

    REGEXP_ID = r"articles\/.*?\/.*?\/(?P<record_id>\d+)(?:\/(?P<version>\d+)|)"

    # the base entry point of the REST API
    API_URL = "https://api.figshare.com/v2"

    # the files and metadata about the dataset
    META_FILES_JSONPATH = "files[*]"

    # paths to file attributes
    ATTR_FILE_LINK_JSONPATH = "download_url"
    ATTR_NAME_JSONPATH = "name"
    ATTR_SIZE_JSONPATH = "size"
    ATTR_HASH_JSONPATH = "computed_md5"
    ATTR_HASH_TYPE_VALUE = "md5"

    @property
    def API_URL_META(self):
        s = "{api_url}/articles/{record_id}"

        if self._params.get("version", None):
            s += "/versions/{version}"

        return s


class DjehutyDataset(FigShareDataset):
    """Downloader for Djehuty repository."""

    REGEXP_ID = r"articles\/.*?\/(?P<record_id>\d+)(?:\/(?P<version>\d+)|)"

    # the base entry point of the REST API
    API_URL = "https://data.4tu.nl/v2"


class GitHubDataset(DatasetDownloader):
    """Downloader for GitHub repository."""

    API_URL = "https://github.com/"
    REGEXP_ID = r"github\.com\/(?P<record_id>[a-zA-Z0-9]+\/[a-zA-Z0-9]+)[\/]*.*"

    def _get(self, output_folder: Union[Path, str], *args, **kwargs):
        res = requests.get(
            f"{self.API_URL}{self._params['record_id']}/archive/refs/heads/master.zip"
        )
        z = zipfile.ZipFile(io.BytesIO(res.content))
        z.extractall(output_folder)

    @property
    def files(self):
        # at the moment, .files is not available for GitHub
        raise NotImplementedError("'files' is not available for GitHub")


class HuggingFaceDataset(DatasetDownloader):
    """Downloader for Huggingface repository."""

    REGEXP_ID = r"huggingface.co/datasets/(?P<record_id>.*)"

    def _get(
        self,
        output_folder: Union[Path, str],
    ):
        try:
            from datasets import load_dataset
        except ImportError as err:
            raise ImportError(
                "Install 'datasets' to use HuggingFace Datasets"
                " or use 'pip install datahugger[all]'"
            ) from err

        params = self.params if self.params else {}
        load_dataset(self._params["record_id"], cache_dir=output_folder, **params)

    @property
    def files(self):
        # at the moment, .files is not available for HuggingFace
        raise NotImplementedError("'files' is not available for HuggingFace")


class MendeleyDataset(DatasetDownloader):
    """Downloader for Mendeley repository."""

    REGEXP_ID = r"data\.mendeley\.com\/datasets\/(?P<record_id>[0-9a-z]+)(?:\/(?P<version>\d+)|)"  # noqa

    # the base entry point of the REST API
    API_URL = "https://data.mendeley.com/public-api/"

    # version url
    API_URL_VERSION = "{api_url}datasets/{record_id}/versions"

    # the files and metadata about the dataset
    API_URL_META = (
        "{api_url}datasets/{record_id}/files?folder_id=root&version={version}"
    )

    # paths to file attributes
    ATTR_FILE_LINK_JSONPATH = "content_details.download_url"
    ATTR_NAME_JSONPATH = "filename"
    ATTR_SIZE_JSONPATH = "size"
    ATTR_HASH_JSONPATH = "content_details.sha256_hash"
    ATTR_HASH_TYPE_VALUE = "sha256"


class OSFDataset(DatasetDownloader):
    """Downloader for OSF repository."""

    REGEXP_ID = r"osf\.io\/(?P<record_id>.*)/"

    # the base entry point of the REST API
    API_URL = "https://api.osf.io/v2/registrations/"

    # the files and metadata about the dataset
    API_URL_META = "{api_url}{record_id}/files/osfstorage/?format=jsonapi"
    META_FILES_JSONPATH = "data[*]"

    PAGINATION_JSONPATH = "links.next"

    # paths to file attributes
    ATTR_KIND_JSONPATH = "attributes.kind"

    ATTR_FILE_LINK_JSONPATH = "links.download"
    ATTR_FOLDER_LINK_JSONPATH = "relationships.files.links.related.href"

    ATTR_NAME_JSONPATH = "attributes.name"
    ATTR_SIZE_JSONPATH = "attributes.size"
    ATTR_HASH_JSONPATH = "attributes.extra.hashes.sha256"
    ATTR_HASH_TYPE_VALUE = "sha256"


class ZenodoDataset(DatasetDownloader):
    """Downloader for Zenodo repository.

    For Zenodo records, new versions have new identifiers.
    """

    REGEXP_ID = r"zenodo\.org\/record\/(?P<record_id>\d+).*"

    # the base entry point of the REST API
    API_URL = "https://zenodo.org/api/"

    # the files and metadata about the dataset
    API_URL_META = "{api_url}records/{record_id}"
    META_FILES_JSONPATH = "files[*]"

    # paths to file attributes
    ATTR_NAME_JSONPATH = "key"
    ATTR_FILE_LINK_JSONPATH = "links.self"
    ATTR_SIZE_JSONPATH = "size"
    ATTR_HASH_JSONPATH = "checksum"

    def _get_attr_hash(self, record):
        return self._get_attr_attr(record, self.ATTR_HASH_JSONPATH).split(":")[1]

    def _get_attr_hash_type(self, record):
        return self._get_attr_attr(record, self.ATTR_HASH_JSONPATH).split(":")[0]



from datahugger.handles import DOI
from datahugger.handles import ArXiv
from datahugger.handles import Handle
from datahugger.handles import is_arxiv
from datahugger.handles import is_doi
from datahugger.handles import is_handle
from datahugger.resolvers import _resolve_service
from datahugger.utils import _is_url


def parse_resource_identifier(resource, resolve=True):
    """Parse resource identifier or location.

    Arguments
    ---------
    resource: str, pathlib.Path
        The URL, DOI, or Handle of the dataset.
    resolve: bool
        Resolve handles (e.g. DOIs and Handles). Default: True.

    Returns
    -------

    str, DOI, Handle, ArXiv
        A parsed and standardised resource handle.
    """

    if isinstance(resource, (DOI, Handle, ArXiv)):
        handle = resource
    elif isinstance(resource, str) and is_doi(resource):
        handle = DOI.parse(resource)

        if resolve:
            handle.resolve()
    elif isinstance(resource, str) and is_handle(resource):
        handle = Handle.parse(resource)

        if resolve:
            handle.resolve()
    elif isinstance(resource, str) and is_arxiv(resource):
        handle = ArXiv.parse(resource)
    elif isinstance(resource, str) and _is_url(resource):
        handle = resource
    else:
        raise ValueError(
            f"'{resource}' is not a correct resource "
            "identifier (e.g. a URL, DOI, Handle)"
        )

    return handle


def info(
    resource,
    max_file_size=None,
    force_download=False,
    unzip=True,
    progress=True,
    print_only=False,
    params=None,
):
    """Get info on the content of the dataset.

    Arguments
    ---------
    resource: str, pathlib.Path
        The URL, DOI, or Handle of the dataset.
    max_file_size: int
        The maximum number of bytes for a single file. If exceeded,
        the file is skipped.
    force_download: bool
        Force the download of the dataset even if there are already
        files in the destination folder. Default: False.
    unzip: bool
        Unzip is the output is a single zip file. Default: True.
    progress: bool
        Print the progress of the download. Default: True.
    print_only: bool
        Print the output of the dataset download without downloading
        the actual files (Dry run). Default: False.
    params: dict
        Extra parameters for the request.

    Returns
    -------

    datahugger.base.DatasetDownloader
        The dataset download object for the specific service.
    """

    handle = parse_resource_identifier(resource)
    service_class = _resolve_service(handle)

    return service_class(
        handle,
        max_file_size=max_file_size,
        force_download=force_download,
        unzip=unzip,
        progress=progress,
        print_only=print_only,
        params=params,
    )


def get(
    resource,
    output_folder,
    max_file_size=None,
    force_download=False,
    unzip=True,
    progress=True,
    print_only=False,
    params=None,
):
    """Get the content of repository.

    Download the content of the dataset to a local folder. Provide a
    URL or DOI to the dataset in the data repository.

    Arguments
    ---------
    resource: str, pathlib.Path
        The URL, DOI, or Handle of the dataset.
    output_folder: str, pathlib.Path
        The folder to download the dataset files to.
    max_file_size: int
        The maximum number of bytes for a single file. If exceeded,
        the file is skipped.
    force_download: bool
        Force the download of the dataset even if there are already
        files in the destination folder. Default: False.
    unzip: bool
        Unzip is the output is a single zip file. Default: True.
    progress: bool
        Print the progress of the download. Default: True.
    print_only: bool
        Print the output of the dataset download without downloading
        the actual files (Dry run). Default: False.
    params: dict
        Extra parameters for the request.

    Returns
    -------

    datahugger.base.DatasetDownloader
        The dataset download object for the specific service.
    """

    service = info(
        resource,
        max_file_size=max_file_size,
        force_download=force_download,
        unzip=unzip,
        progress=progress,
        print_only=print_only,
        params=params,
    )

    return service.download(output_folder)



import pytest

import datahugger


@pytest.mark.parametrize(
    "url_or_id",
    [
        ("https://doi.org/10.5281/zenodo.6614829"),
    ],
)
def test_load_zenodo_6614829(url_or_id, tmpdir, capsys):
    dh = datahugger.get(url_or_id, tmpdir, max_file_size=1e6)

    dh.tree()

    captured = capsys.readouterr()
    assert "quasiperiod.m" in captured.out

    # test count
    assert "12 file" in captured.out
    assert 12 == (len(dh) - 1)


def test_load_github_cbsodata(tmpdir, capsys):
    dh = datahugger.get("https://github.com/j535d165/cbsodata", tmpdir)

    dh.tree()

    captured = capsys.readouterr()
    assert ".gitignore" in captured.out



import requests


class MetaData:
    """MetaData for a resource"""

    def __init__(self, resource):
        self.resource = resource

    def _get_doi_metadata(self, content_type):
        r = requests.get(
            f"https://doi.org/{self.resource}",
            headers={"Accept": content_type},
        )
        r.raise_for_status()

        return r

    def rdf(self):
        """Return metadata in RDF representation"""
        return self._get_doi_metadata("application/rdf+xml").text

    def cls(self):
        """Return metadata in CLS representation"""
        return self._get_doi_metadata("application/vnd.citationstyles.csl+json").json()

    def bibtex(self):
        """Return metadata in BibTex representation"""
        return self._get_doi_metadata("application/x-bibtex").text

    def ris(self):
        """Return metadata in RIS representation"""
        return self._get_doi_metadata("application/x-research-info-systems").text

    def citation(self, style=None, locale=None):
        """Return metadata as formatted citation.

        See https://editor.citationstyles.org/searchByName/ for
        an overview of styles.
        """
        content_type = "text/x-bibliography"

        if style:
            content_type += f"; style={style}"
        if locale:
            content_type += f"; locale={locale}"

        return self._get_doi_metadata(content_type).text



from datahugger.api import get
from datahugger.api import info
from datahugger.api import parse_resource_identifier
from datahugger.base import DownloadResult
from datahugger.exceptions import DOIError
from datahugger.exceptions import RepositoryNotSupportedError

__all__ = [
    "get",
    "info",
    "parse_resource_identifier",
    "DownloadResult",
    "DOIError",
    "RepositoryNotSupportedError",
]

try:
    from datahugger._version import __version__
    from datahugger._version import __version_tuple__
except ImportError:
    __version__ = "0.0.0"
    __version_tuple__ = (0, 0, 0)



import argparse
import datetime
import json
from pathlib import Path

import pandas as pd
import requests

# from pydatacite import DOIs
import datahugger
from datahugger import RepositoryNotSupportedError

BENCHMARK_FILE = Path("benchmark", "benchmark_datasets.csv")
BENCHMARK_OUTPUT_FILE = Path("benchmark", "benchmark_datasets_tested.csv")


def create_dataset(args):
    records = [
        x
        for y in range(1)
        for x in requests.get(
            "https://api.datacite.org/dois?random=true&page[size]=100"
        ).json()["data"]
    ]

    result = []
    for r in records:
        result.append({"id": r["id"], "type": r["type"], "url": r["attributes"]["url"]})

    pd.DataFrame(result).to_csv(
        Path("benchmark", "datacite", f"datacite_random_{datetime.datetime.now()}"),
        index=False,
    )


def merge_datasets(args):
    datasets = Path("benchmark", "datacite").glob("*")

    df = pd.concat([pd.read_csv(d) for d in datasets], axis=0)
    df.drop_duplicates(inplace=True)

    df.to_csv(BENCHMARK_FILE, index=False)


def run(args):
    df = pd.read_csv(BENCHMARK_FILE)

    df[["service", "error"]] = df.apply(test_repo, axis=1, result_type="expand")
    df.to_csv(BENCHMARK_OUTPUT_FILE)

    print(df)


def test_repo(r):
    try:
        cl = datahugger.info(r["id"]).__class__.__name__
        print(r["id"], f"service found: {cl}")
        return {"service": cl, "error": None}
    except RepositoryNotSupportedError:
        # Repository not supported, but no error
        return {"service": None, "error": None}
    except Exception as err:
        print(r["id"], f"results in: {err}")
        return {"service": None, "error": str(err)}


def get_coverage(args):
    df = pd.read_csv(BENCHMARK_OUTPUT_FILE)
    cov = df["service"].notnull().sum() / len(df)

    print(f"Dataset coverage: {cov * 100}%")

    with open(".datacite_coverage.json", "w") as f:
        json.dump({"datasets": cov * 100}, f)


def get_report(args):
    df = pd.read_csv(BENCHMARK_OUTPUT_FILE, index_col=0)

    frac_supported = df["service"].notnull().sum() / len(df)
    frac_with_error = df["error"].notnull().sum() / len(df)
    frac_not_supported = 1 - frac_supported - frac_with_error

    print("## Coverage report")
    print(
        f"The following benchmark was applied to {len(df)}",
        "randomly selected records from Datacite.",
    )
    print()
    print("### Percentages")
    print(f"Percentage of datasets supported: {frac_supported*100:.1f}%")
    print(f"Percentage of datasets not supported: {frac_not_supported*100:.1f}%")
    print(f"Percentage of datasets with error: {frac_with_error*100:.1f}%")

    print()
    print("### Table with unexpected errors")
    print(df[df["error"].notnull()].to_markdown())


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(help="Coverage tool")

    parser_create = subparsers.add_parser("create_dataset", help="Create dataset")
    parser_create.set_defaults(func=create_dataset)

    parser_merge = subparsers.add_parser("merge_dataset", help="merge dataset")
    parser_merge.set_defaults(func=merge_datasets)

    parser_run = subparsers.add_parser("run", help="Run coverage on dataset")
    parser_run.set_defaults(func=run)

    parser_coverage = subparsers.add_parser("coverage", help="Coverage on dataset")
    parser_coverage.set_defaults(func=get_coverage)

    parser_coverage = subparsers.add_parser("report", help="Report on dataset")
    parser_coverage.set_defaults(func=get_report)

    args = parser.parse_args()
    args.func(args)



import pandas as pd
import requests

url_datacite_clients = (
    "https://api.datacite.org/clients?page%5Bnumber%5D=1&page%5Bsize%5D=1000"
)

data = []
while url_datacite_clients:
    r = requests.get(url_datacite_clients)
    res = r.json()
    data.extend(res["data"])
    url_datacite_clients = res["links"]["next"] if "next" in res["links"] else None


print(len(data))

result = []

for client in data:
    result.append(
        {
            "datacite_id": client["id"],
            "name": client["attributes"]["name"],
            "url": client["attributes"]["url"],
            "re3data": client["attributes"]["re3data"],
        }
    )

df = pd.DataFrame(result)

print(df)
breakpoint()



import logging
import re

import requests

from datahugger.exceptions import DOIError
from datahugger.metadata import MetaData
from datahugger.utils import _is_url

__all__ = ["ArXiv", "DOI", "Handle", "is_doi", "is_handle", "is_arxiv"]


class DOI:
    """DOI class"""

    def __init__(self, doi):
        super().__init__()
        self.doi = doi

    def __str__(self):
        return f"{self.doi}"

    @classmethod
    def parse(cls, doi_str):
        if doi_str.startswith("doi:"):
            doi_str = doi_str[4:]

        if _is_url(doi_str) and "doi.org" in doi_str:
            doi_str = doi_str.split("doi.org/")[1]

        # Thanks to Andrew Gilmartin
        # https://www.crossref.org/blog/dois-and-matching-regular-expressions/
        match = re.match(r"^10.\d{4,9}/[-._;()/:A-Z0-9]+$", doi_str, re.IGNORECASE)
        if not (match is not None and match.group() is not None):
            raise ValueError("Not a valid DOI")

        return cls(match.group())

    @property
    def url(self):
        """Return the url."""

        if hasattr(self, "_resolved_url"):
            return self._resolved_url

        return None

    def resolve(self):
        if hasattr(self, "_resolved_url"):
            return self._resolved_url

        url = f"https://doi.org/{self.doi}"
        r = requests.head(url, allow_redirects=True, timeout=(3, 10))

        if r.status_code == 404 and r.url and r.url.startswith("https://doi.org"):
            raise DOIError(f"DOI {self.doi} not found in the DOI system")
        elif r.status_code in [404, 405]:
            # head request not allowed or possible, try get request
            r = requests.get(url, allow_redirects=True, timeout=(3, 10))
        elif r.status_code in [403]:
            # Most likely a service that tries to prevent webscraping.
            # Might still have an API, so forwaring the response url.
            pass
        else:
            r.raise_for_status()

        logging.info(f"Redirect from {url} to {r.url}")
        self._resolved_url = r.url

        return self._resolved_url

    @property
    def metadata(self):
        if hasattr(self, "_metadata_obj"):
            return self._metadata_obj

        self._metadata_obj = MetaData(self.doi)

        return self._metadata_obj


def is_doi(s: str) -> bool:
    """Check if string is DOI.

    Parameters
    ----------
    s: str
        The string to check for DOI

    Returns
    -------
    bool:
        Is the string a pure DOI or not.
    """

    try:
        DOI.parse(s)
        return True
    except ValueError:
        return False


class Handle:
    """docstring for Handle"""

    def __init__(self, handle):
        super().__init__()
        self.handle = handle

    def __str__(self):
        return f"{self.handle}"

    @classmethod
    def parse(cls, s):
        if s.startswith("hdl:"):
            s = s[4:]

        match = re.match(r"^https\:\/\/hdl\.handle\.net\/(.*)$", s, re.IGNORECASE)
        if not (match is not None and match.group() is not None):
            raise ValueError("Not a valid Handle")

        return cls(match.group())

    @property
    def url(self):
        """Return the url."""

        if hasattr(self, "_resolved_url"):
            return self._resolved_url

        return None

    def resolve(self):
        if hasattr(self, "_resolved_url"):
            return self._resolved_url

        url = f"https://hdl.handle.net/{self.handle}"
        r = requests.head(url, allow_redirects=True, timeout=(3, 10))

        if r.status_code == 404 and r.url and r.url.startswith("https://handle.org"):
            raise ValueError(f"Handle {self.handle} not found in the Handle system")
        elif r.status_code in [404, 405]:
            # head request not allowed or possible, try get request
            r = requests.get(url, allow_redirects=True, timeout=(3, 10))
        elif r.status_code in [403]:
            # Most likely a service that tries to prevent webscraping.
            # Might still have an API, so forwaring the response url.
            pass
        else:
            r.raise_for_status()

        logging.info(f"Redirect from {url} to {r.url}")
        self._resolved_url = r.url

        return self._resolved_url


def is_handle(s: str) -> bool:
    """Check if string is Handle.

    Parameters
    ----------
    s: str
        The string to check for Handle

    Returns
    -------
    bool:
        Is the string a pure Handle or not.
    """

    try:
        Handle.parse(s)
        return True
    except ValueError:
        return False


class ArXiv:
    """docstring for arXiv"""

    def __init__(self, handle, version=None):
        super().__init__()
        self.handle = handle
        self.version = version

    def __str__(self):
        return f"{self.handle}"

    @classmethod
    def parse(cls, s):
        match = re.match(
            r".*(arxiv\.org/abs/|arXiv:)(\d{4}.\d{4,5}|[a-z\-]+(\.[A-Z]{2})?\/\d{7})(v\d+)?$",
            s,
            re.IGNORECASE,
        )

        if not (match is not None and match.group() is not None):
            raise ValueError("Not a valid arXiv identifier")

        return cls(match.group(2))

    @property
    def url(self):
        """Return the url."""

        return f"https://arxiv.org/abs/{self.handle}"

    @property
    def metadata(self):
        if hasattr(self, "_metadata_obj"):
            return self._metadata_obj

        self._metadata_obj = MetaData(f"https://doi.org/10.48550/arXiv.{self.handle}")
        return self._metadata_obj


def is_arxiv(s: str) -> bool:
    """Check if string is arXiv.

    Parameters
    ----------
    s: str
        The string to check for arXiv

    Returns
    -------
    bool:
        Is the string a pure arXiv or not.
    """

    try:
        ArXiv.parse(s)
        return True
    except ValueError:
        return False



import io
import logging
import os
import re
import zipfile
from pathlib import Path
from typing import Union
from urllib.parse import urlparse

import requests
from jsonpath_ng import parse
from scitree import scitree
from tqdm import tqdm

from datahugger.utils import _format_filename
from datahugger.utils import _get_url
from datahugger.utils import _is_url


class DownloadResult:
    """Result class after downloading the dataset."""

    def __init__(self, dataset, output_folder):
        self.dataset = dataset
        self.output_folder = output_folder

    def __str__(self):
        return f"<{self.__class__.__name__} n_files={len(self)} >"

    def __len__(self):
        return len(self.dataset.files)

    def tree(self, **kwargs):
        """Return the folder tree.

        Tree based on scientific sort.
        """

        return scitree(self.output_folder, **kwargs)


class DatasetDownloader:
    """Base class for downloading resources from repositories."""

    API_URL = None

    def __init__(
        self,
        resource,
        max_file_size=None,
        force_download=False,
        progress=True,
        unzip=True,
        print_only=False,
        params=None,
    ):
        super().__init__()
        self.resource = resource
        self.max_file_size = max_file_size
        self.force_download = force_download
        self.progress = progress
        self.unzip = unzip
        self.print_only = print_only
        self.params = params

    def _get_attr_attr(self, record, jsonp):
        try:
            jsonpath_expression = parse(jsonp)
            return jsonpath_expression.find(record)[0].value
        except Exception:
            return None

    def _get_attr_link(self, record, **kwargs):
        # get the link to the folder
        if self._get_attr_kind(record) == "folder":
            if not hasattr(self, "ATTR_FOLDER_LINK_JSONPATH"):
                return None

            return self._get_attr_attr(record, self.ATTR_FOLDER_LINK_JSONPATH)

        # get the link to the file
        else:
            if not hasattr(self, "ATTR_FILE_LINK_JSONPATH"):
                return None

            return self._get_attr_attr(record, self.ATTR_FILE_LINK_JSONPATH)

    def _get_attr_name(self, record):
        if not hasattr(self, "ATTR_NAME_JSONPATH"):
            return None

        return self._get_attr_attr(record, self.ATTR_NAME_JSONPATH)

    def _get_attr_size(self, record):
        if not hasattr(self, "ATTR_SIZE_JSONPATH"):
            return None

        return self._get_attr_attr(record, self.ATTR_SIZE_JSONPATH)

    def _get_attr_hash(self, record):
        if not hasattr(self, "ATTR_HASH_JSONPATH"):
            return None

        return self._get_attr_attr(record, self.ATTR_HASH_JSONPATH)

    def _get_attr_hash_type(self, record):
        if hasattr(self, "ATTR_HASH_TYPE_VALUE"):
            return self.ATTR_HASH_TYPE_VALUE

        if not hasattr(self, "ATTR_HASH_TYPE_JSONPATH"):
            return None

        return self._get_attr_attr(record, self.ATTR_HASH_TYPE_JSONPATH)

    def _get_attr_kind(self, record):
        if not hasattr(self, "ATTR_KIND_JSONPATH"):
            return "file"

        return self._get_attr_attr(record, self.ATTR_KIND_JSONPATH)

    def download_file(
        self,
        file_link,
        output_folder,
        file_name,
        file_size=None,
        file_hash=None,
        file_hash_type=None,
    ):
        """Download a single file.

        Arguments
        ---------
        file_link: str
            Path to the file to download.
        output_folder: str
            The folder to store the downloaded file.
        file_name: str
            The filename of the downloaded file.
        file_size: int
            The size of the file in bytes.
        file_hash: str
            The MD5 hash of the file.

        """
        if (
            file_size is not None
            and self.max_file_size is not None
            and file_size >= self.max_file_size
        ):
            logging.info(f"Skipping large file {file_link}")
            if self.progress:
                print(f"{_format_filename(file_name)}: SKIPPED")
            return

        if not self.print_only:
            logging.info(f"Downloading file {file_link}")
            res = requests.get(file_link, stream=True)

            output_fp = Path(output_folder, file_name)
            Path(output_fp).parent.mkdir(parents=True, exist_ok=True)

            if not self.force_download and output_fp.exists():
                print("File already exists:", file_name)
                return

            if self.progress:
                with tqdm.wrapattr(
                    open(output_fp, "wb"),
                    "write",
                    miniters=1,
                    desc=_format_filename(file_name),
                    total=int(res.headers.get("content-length", 0)),
                    bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt}",
                ) as fout:
                    for chunk in res.iter_content(chunk_size=4096):
                        fout.write(chunk)
            else:
                with open(output_fp, "wb") as f:
                    f.write(res.content)
        else:
            print(f"{_format_filename(file_name)}: COMPLETE")

    def _parse_url(self, url):
        if not isinstance(url, str) or not _is_url(url):
            raise ValueError("Not a valid URL.")

        try:
            match = re.search(self.REGEXP_ID, url)
            return match.groupdict()
        except Exception as err:
            raise ValueError(f"Failed to parse URL '{url}'") from err

    def _unpack_single_folder(self, zip_url, output_folder):
        r = requests.get(zip_url)
        z = zipfile.ZipFile(io.BytesIO(r.content))

        for zip_info in z.infolist():
            if zip_info.filename[-1] == "/":
                continue
            zip_info.filename = os.path.basename(zip_info.filename)
            z.extract(zip_info, output_folder)

    def _pre_files(self):
        pass

    def _get_files_recursive(self, url, folder_name=None, base_url=None):
        if not isinstance(url, str):
            ValueError(f"Expected url to be string type, got {type(url)}")

        result = []

        # get the data from URL
        res = requests.get(url)
        response = res.json()

        # find path to raw files
        if hasattr(self, "META_FILES_JSONPATH"):
            if isinstance(self.META_FILES_JSONPATH, str):
                jsonpath_expression = parse(self.META_FILES_JSONPATH)
            else:
                jsonpath_expression = self.META_FILES_JSONPATH

            files_raw = [x.value for x in jsonpath_expression.find(response)]
        else:
            files_raw = response

        for f in files_raw:
            # create the file or folder path
            if folder_name is None:
                f_path = self._get_attr_name(f)
            else:
                f_path = str(Path(folder_name, self._get_attr_name(f)))

            if self._get_attr_kind(f) == "folder":
                result.extend(
                    self._get_files_recursive(
                        self._get_attr_link(f, base_url=base_url), folder_name=f_path
                    )
                )
            else:
                result.append(
                    {
                        "link": self._get_attr_link(f, base_url=base_url),
                        "name": f_path,
                        "size": self._get_attr_size(f),
                        "hash": self._get_attr_hash(f),
                        "hash_type": self._get_attr_hash_type(f),
                    }
                )

        if hasattr(self, "PAGINATION_JSONPATH"):
            jsonpath_expression = parse(self.PAGINATION_JSONPATH)
            next_url = jsonpath_expression.find(response)[0].value

            if next_url:
                result.extend(
                    self._get_files_recursive(next_url, folder_name=folder_name)
                )

        return result

    @property
    def _params(self):
        """Params including url params."""
        if hasattr(self, "__params"):
            return self.__params

        url = _get_url(self.resource)
        url_params = self._parse_url(url)
        if self.params:
            new_params = self.params.copy()
            new_params.update(url_params)
            self.__params = new_params
        else:
            self.__params = url_params

        return self.__params

    @property
    def files(self):
        if hasattr(self, "_files"):
            return self._files

        self._pre_files()

        url = _get_url(self.resource)
        uri = urlparse(url)
        base_url = uri.scheme + "://" + uri.netloc

        self._files = self._get_files_recursive(
            self.API_URL_META.format(
                api_url=self.API_URL, base_url=base_url, **self._params
            ),
            base_url=base_url,
        )

        return self._files

    def _get(
        self,
        output_folder: Union[Path, str],
    ):
        if (
            len(self.files) == 1
            and (
                self.files[0]["link"].endswith(".zip")
                or self.files[0]["name"].endswith(".zip")
            )
            and self.unzip
        ):
            self._unpack_single_folder(self.files[0]["link"], output_folder)
            return

        for f in self.files:
            self.download_file(
                f["link"],
                output_folder,
                file_name=f["name"],
                file_size=f["size"],
                file_hash=f["hash"],
                file_hash_type=f["hash_type"],
            )

    def download(
        self,
        output_folder: Union[Path, str],
    ):
        """Download files.

        Arguments
        ---------
        output_folder: str
            The folder to store the downloaded results.

        """

        self._get(output_folder=output_folder)

        return DownloadResult(self, output_folder)



import logging
import re
from urllib.parse import urlparse

import requests

from datahugger.config import RE3DATA_SOFTWARE
from datahugger.config import SERVICES_NETLOC
from datahugger.config import SERVICES_NETLOC_REGEXP
from datahugger.exceptions import RepositoryNotSupportedError
from datahugger.handles import DOI
from datahugger.utils import _get_url
from datahugger.utils import get_datapublisher_from_doi
from datahugger.utils import get_re3data_repositories
from datahugger.utils import get_re3data_repository


def _resolve_service(resource):
    for resolver in [
        _resolve_service_from_netloc,
        _resolve_service_from_url_pattern,
        _resolve_service_with_re3data,
    ]:
        service_class = resolver(resource)

        if service_class is not None:
            logging.info(f"Service found: {service_class}")
            return service_class

    raise RepositoryNotSupportedError(f"Data protocol for {resource} not found.")


def _resolve_service_from_netloc(resource):
    uri = urlparse(_get_url(resource))

    if not uri.hostname:
        return None

    logging.info(f"Resolve service for netloc '{uri.hostname}'")
    if uri.hostname in SERVICES_NETLOC.keys():
        return SERVICES_NETLOC[uri.hostname]


def _resolve_service_from_url_pattern(resource):
    url = _get_url(resource)

    for netloc_re, service in SERVICES_NETLOC_REGEXP.items():
        if re.match(netloc_re, url):
            return service


def _resolve_service_with_re3data(doi):
    if not isinstance(doi, DOI):
        return None

    logging.info("Resolve service with datacite and re3data")
    try:
        publisher = get_datapublisher_from_doi(doi)
    except requests.HTTPError:
        return None
    logging.info(f"Datacite publisher of dataset: {publisher}")

    if not publisher:
        logging.info("Can't resolve the publisher from the DOI.")
        return None

    data_repos = get_re3data_repositories()

    for repo in data_repos:
        if publisher.lower() == repo["name"].lower():
            r_software = get_re3data_repository(repo["id"])

            try:
                return RE3DATA_SOFTWARE[r_software.lower()]
            except KeyError:
                return None



import re
from pprint import pprint

from py3data import Repositories


def count_repos():
    counts = {
        "dataverse": Repositories().filter(software="Dataverse").count(),
        "dspace": Repositories().filter(software="DSpace").count(),
        "figshare": Repositories().query("figshare").count(),
        "dataone": Repositories().query("dataone").count(),
        "zenodo": 1,
        "mendeley": 1,
        "osf": 1,
        "dryad": 1,
        "github": 1,
        "huggingface": 1,
        "pangaea": 1,
    }

    print(counts)

    return sum([v for k, v in counts.items()])


def _update_docs(file_name, n_total):
    with open(file_name) as f_read:
        readme = f_read.read()

    print(readme)

    readme_updated = re.sub(
        r"\<\!\-\-\scount\s\-\-\>(\d+)\<\!\-\-\scount\s\-\-\>",
        f"<!-- count -->{n_total}<!-- count -->",
        readme,
        flags=re.MULTILINE,
    )

    with open(file_name, "w") as f_write:
        f_write.write(readme_updated)


if __name__ == "__main__":
    n_total = count_repos()
    pprint(n_total)
    print("Number of supported data repositories", n_total)

    _update_docs("README.md", n_total)
    _update_docs("docs/repositories.md", n_total)



import pytest

import datahugger
from datahugger.api import _resolve_service
from datahugger.handles import DOI
from datahugger.handles import ArXiv
from datahugger.services import DataverseDataset


def test_resolve_service():
    doi = DOI.parse("10.34894/FXUGHW")
    doi.resolve()

    service = _resolve_service(doi)(doi)

    assert isinstance(service, DataverseDataset)


@pytest.mark.xfail(raises=ValueError)
def test_doi_not_found(tmpdir):
    doi = "10.1038/s42256-020-00287"

    datahugger.get(doi, tmpdir)


def test_resolve_service_via_doi_handle(tmpdir):
    doi = DOI.parse("10.34894/FXUGHW")
    doi.resolve()

    r = datahugger.get(doi, tmpdir)

    assert isinstance(doi, DOI)
    assert isinstance(r.dataset, DataverseDataset)


def test_get_doi_metadata_cls(tmpdir):
    doi = DOI.parse("10.34894/FXUGHW")
    doi.resolve()

    m = doi.metadata.cls()

    assert isinstance(m, dict)


def test_get_doi_metadata_bibtex(tmpdir):
    doi = DOI.parse("10.34894/FXUGHW")
    doi.resolve()

    m = doi.metadata.bibtex()

    assert isinstance(m, str)


def test_arxiv_handle(tmpdir):
    arxiv = ArXiv.parse("https://arxiv.org/abs/astro-ph/9802301v1")

    m = arxiv.metadata.cls()

    assert isinstance(m, dict)


def test_get_doi_metadata_from_instance(tmpdir):
    # initiate datahugger downloader and then get metadata
    service = datahugger.info("10.34894/FXUGHW")
    m1 = service.resource.metadata.citation()

    # initiate datahugger resource and get metadata
    doi = DOI.parse("10.34894/FXUGHW")
    doi.resolve()
    m2 = doi.metadata.citation()

    assert m1 == m2 is not None
